def dictify(h):

    def next2yield(h):
        while True:
            n = h.next()
            if not n:
                break
            yield n

    result = {}
    for key in next2yield(h.keys()):
        if hasattr(h, 'h'):
            # haxe version change
            result[key] = h.h.get(key)
        else:
            result[key] = h.get(key)

    return result



import daff
import sqlalchemy
from sqlalchemy import create_engine
from sqlalchemy.orm import create_session
from sqlalchemy.exc import CompileError
from sqlalchemy.ext.declarative import declarative_base

from catsql.database import Database
from catsql.daffsql.sqlalchemy_helper import SqlAlchemyHelper


class SqlAlchemyDatabase(daff.SqlDatabase):

    def __init__(self, url):
        if isinstance(url, Database):
            db = url
            self.database = db
            self.url = db.url
            self.Base = db.Base
            self.engine = db.engine
            self.session = db._session
        else:
            self.database = None
            self.url = url
            self.Base = declarative_base()
            self.engine = create_engine(url)
            self.Base.metadata.reflect(self.engine)
            self.session = create_session(bind=self.engine)
        self.helper = SqlAlchemyHelper()

    def finalize(self):
        if self.database:
            changes = (self.helper.updates +
                       self.helper.inserts +
                       self.helper.deletes)
            self.database.finalize(changes > 0)

    def getTable(self, name):
        self.getColumns(name)
        result = self.Base.metadata.tables[name.toString()]
        return result

    @property
    def events(self):
        return {
            'updates': self.helper.updates,
            'inserts': self.helper.inserts,
            'deletes': self.helper.deletes,
            'skips': self.helper.skips
        }

    # needed because pragmas do not support bound parameters
    def getQuotedColumnName(self, name):
        return name  # adequate for test, not real life

    # needed because pragmas do not support bound parameters
    def getQuotedTableName(self, name):
        return name.toString()  # adequate for test, not real life

    def getColumns(self, name):
        name = name.toString()
        tab = self.Base.metadata.tables[name]
        columns = []
        for name, col in tab.columns.items():
            column = daff.SqlColumn()
            column.setName(name)
            column.setPrimaryKey(col.primary_key)
            try:
                type_name = str(col.type)
                if isinstance(col.type, sqlalchemy.types.Float):
                    type_name = 'REAL'
                column.setType(type_name, 'sqlalchemy')
            except CompileError:
                column.setType("", 'sqlalchemy')
            columns.append(column)
        return columns

    def begin(self, query, args=[], order=[]):
        print("Not implemented " + query)
        return False

    def beginRow(self, tab, row, order=[]):
        print("Not implemented")
        return False

    def read(self):
        return False

    def get(self, index):
        return None

    def end(self):
        pass

    def rowid(self):
        return "rowid"

    def getHelper(self):
        return self.helper

    def getNameForAttachment(self):
        return None



def add_options(parser):

    parser.add_argument('catsql_database_url',
                        help='Database url or filename.  Examples: '
                        'sqlite:///data.db, '
                        'mysql://user:pass@host/db, '
                        'postgres[ql]://user:pass@host/db, '
                        'data.sqlite3')

    parser.add_argument('--column', action='append',
                        help='Column to include (defaults to all columns). '
                        'Can be a comma separated list of multiple columns.')

    parser.add_argument('--count', default=False, action='store_true',
                        help='Show row counts instead of actual data.')

    parser.add_argument('--csv', default=False, action='store_true',
                        help='Output strictly in CSV format. Only one table can be shown.')

    parser.add_argument('--distinct', default=False, action='store_true',
                        help='Show distinct rows only, hiding duplicates.')

    parser.add_argument('--edit', required=False, action='store_true',
                        help='Edit original table in your favorite editor. '
                        'Respects $EDITOR environment variable.')

    parser.add_argument('--excel', nargs=1, required=False, default=None,
                        help='Save results to an excel file.')

    parser.add_argument('--grep', action='append',
                        help='Search cells for occurrence of a text fragment. '
                        'Translated to SQL query, performed by database.')

    parser.add_argument('--json', nargs=1, required=False, default=None,
                        help='Save results to a json file. Only one table allowed.')

    parser.add_argument('--limit', nargs=1, required=False, default=None,
                        help='Maximum number of rows per table.')

    parser.add_argument('--load-bookmark', required=False, action='store_true',
                        help='Load a set of filters from a file.')

    parser.add_argument('--order', action='append',
                        help='Columns to order by. '
                        'Can be a comma separated list of columns names. '
                        'Add + or - to end of name to specify ascending or descending '
                        'order.  Specify "none" to disable ordering completely '
                        '(by default we always try to apply some order)')

    parser.add_argument('--output', nargs=1, required=False, default=None,
                        help='Save output to specified file.  Incompatible with --edit.')

    parser.add_argument('--safe-null', required=False, action='store_true',
                        help='Encode nulls in a reversible way.')

    parser.add_argument('--save-bookmark', nargs=1, required=False, default=None,
                        help='Save the current set of filters specified to a file.')

    parser.add_argument('--select-from', nargs=1, required=False, default=None,
                        help='Supply full SQL: "select ... from ..." etc.')

    parser.add_argument('--sql', action='append',
                        help='Add a raw SQL filter for rows to include.  Example: '
                        '"total < 1000", "created_at > now() - interval \'1 day\'". '
                        'Tables that don\'t have the columns mentioned are '
                        'omitted.'
    )

    parser.add_argument('--sqlite', nargs=1, required=False, default=None,
                        help='Save results to a sqlite file.')

    parser.add_argument('--table', action='append',
                        help='Table to include (defaults to all tables). '
                        'Can be a comma separated list of multiple tables.')

    parser.add_argument('--schema', required=False, default=None,
                        help="Database schema to use (default: public).")

    parser.add_argument('--terse', default=False, action='store_true',
                        help='Hide any columns with predetermined values.')

    parser.add_argument('--types', default=False, action='store_true',
                        help='Show column types instead of actual data.')

    parser.add_argument('--value', action='append',
                        help='R|Add a column=value filter. Example:\n'
                        '  --value id=ID --value name=Jupiter\n'
                        'As a shortcut you can also do:\n'
                        '  --id ID --name Jupiter')

    parser.add_argument('--verbose', default=False, action='store_true',
                        help='Show raw SQL queries as they are made.')

    parser.add_argument('--quiet', default=False, action='store_true',
                        help='Omit any optional output to console.')



from catsql.filter import Filter
import os
from shutil import copyfile
from sqlalchemy import Column, create_engine, Integer, MetaData, String, Table
from sqlalchemy.exc import ArgumentError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import create_session, mapper
import sys

if sys.version_info[0] == 2:
    import unicodecsv as csv
else:
    import csv


class Database(object):
    def __init__(self, url, verbose=False, tables=None, schema=None, can_create=False):
        self.url = url
        self._full_url = self.url
        self.verbose = verbose
        self.tables = tables
        self.schema = schema
        self.can_create = can_create
        self.csv = None
        self.table = None
        self.connect_database()

    def finalize(self, changed):
        if not changed:
            # no changes
            return
        if not self.csv:
            # changes already applied externally
            return
        self.save_csv(self.csv)

    def connect_database(self):
        meta = MetaData()
        if self.schema:
            meta.schema = self.schema
        self.Base = declarative_base(metadata=meta)
        if '://' not in self.url and self.url.lower().endswith('.csv'):
            # special case for csv file - just load into memory
            self.engine = self.wrap_csv(self.url)
        else:
            try:
                self.engine = create_engine(self.url, echo=self.verbose)
            except ImportError as e:
                print("Support library for this database not installed - {}".format(e))
                exit(1)
            except ArgumentError:
                if os.path.exists(self.url) or self.can_create:
                    try:
                        # maybe this is a local sqlite database?
                        sqlite_url = 'sqlite:///{}'.format(self.url)
                        self.engine = create_engine(sqlite_url, echo=self.verbose)
                        self._full_url = sqlite_url
                    except ArgumentError:
                        # no joy, recreate the original problem and die.
                        self.engine = create_engine(self.url, echo=self.verbose)
                else:
                    raise
        only = None
        if self.tables:
            only = list(self.tables)
        self.Base.metadata.reflect(self.engine, only=only)
        self._session = create_session(bind=self.engine)

    def wrap_csv(self, url):
        engine = create_engine('sqlite://')
        self.csv = url
        with open(url) as f:
            metadata = MetaData(bind=engine)
            cf = csv.DictReader(f, delimiter=',')
            names = dict([(name, name or ('col%d' % i)) for i, name in enumerate(cf.fieldnames)])
            cols = [Column(names[rowname], String(), primary_key=(rowname == 'id'))
                    for rowname in cf.fieldnames]
            has_primary_key = any((rowname == 'id') for rowname in cf.fieldnames)
            if not has_primary_key:
                cols = [Column('id', Integer(), primary_key=True)] + cols
            table = Table('_table_', metadata, *cols)
            table.create()
            for idx, row in enumerate(cf):
                row = dict((names[name], val) for name, val in row.items())
                if not has_primary_key:
                    row['id'] = idx + 1
                table.insert().values(**row).execute()

            class CsvTable(object):
                pass
            mapper(CsvTable, table)
            self.table = table
        return engine

    def save_csv(self, fname):
        if os.path.exists(fname):
            copyfile(fname, '{}.bak'.format(fname))
        with open(fname, 'wb') as fout:
            writer = csv.writer(fout)
            writer.writerow([c.name for c in self.table.columns][1:])
            for row in self.table.select().order_by('ROWID').execute():
                writer.writerow(row[1:])

    @property
    def session(self):
        return self._session

    @property
    def full_url(self):
        return self._full_url

    @property
    def tables_metadata(self):
        return self.Base.metadata.tables

    def query(self, columns=None):
        return Filter(self, columns=columns)
